# Plugin for TWiki Collaboration Platform, http://TWiki.org/
#
# MasonPlugin:
# 	Copyright (C) 2005-2006 Jos Gilmar Alves Santos Junior, gilmarsantosjr@safernet.org.br
#
# EmptyPlugin:
# 	Copyright (C) 2000-2003 Andrea Sterbini, a.sterbini@flashnet.it
# 	Copyright (C) 2001-2004 Peter Thoeny, peter@thoeny.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
#
# =========================

package TWiki::Plugins::MasonPlugin;

use strict;

use HTML::Mason;
use TWiki::Plugins::MasonPlugin::SessionTieHash;
use Safe;

# =========================
our ($web, $topic, $user, $installWeb, $VERSION, $pluginName, $debug);
our ($parsed, %app_list, %app_err_msg, @found_apps, %comps, $executed);

$VERSION = '1.000';
$pluginName = 'MasonPlugin';
$parsed = 0;
%app_list = ();
%app_err_msg = ();
@found_apps = ();

# =========================
sub initPlugin
{
    ( $topic, $web, $user, $installWeb ) = @_;

    # check for Plugins.pm versions
    if( $TWiki::Plugins::VERSION < 1.1 ) {
        TWiki::Func::writeWarning( "Version mismatch between $pluginName and Plugins.pm" );
        return 0;
    }

    # Get plugin debug flag
    $debug = TWiki::Func::getPluginPreferencesFlag( "DEBUG" );
	%comps = ();
	$executed = 0;

    # Get plugin preferences, the variable defined by:          * Set EXAMPLE = ...
    #$exampleCfgVar = TWiki::Func::getPluginPreferencesValue( "EXAMPLE" ) || "default";
	TWiki::Func::registerTagHandler('MASON',\&MASON_Handler,'classic');
	TWiki::Func::registerTagHandler('MASON_APPS',\&MASON_APPS_Handler,'classic');
	TWiki::Func::registerTagHandler('MASON_ENABLED_APPS',\&MASON_ENABLED_APPS_Handler,'classic');
	TWiki::Func::registerTagHandler('MASON_APPS_TABLE',\&MASON_APPS_TABLE_Handler,'classic');
	TWiki::Func::registerTagHandler('MASON_FAILED_APPS',\&MASON_FAILED_APPS_Handler,'classic');
    
	# Plugin correctly initialized
    TWiki::Func::writeDebug( "- TWiki::Plugins::${pluginName}::initPlugin( $web.$topic ) is OK" ) if $debug;
    return 1;
}

=pod

---++ commonTagsHandler($text, $topic, $web)

Replace all mason  component calls  with the due result. It's done
here cause there must be two passes: the first collect information 
about all  component calls to  build a  single  component for each
application,  and this  replace the  calls with  the results. This
way it's only a single execution per application is needed.

=cut

sub commonTagsHandler
{
	# my ($text, $topic, $web) = @_;
	$_[0] =~ s/%MASON_PLUGIN_RENDERED{app="(.*?)" call_id="(.*?)"}%/&_result($1, $2)/ge;
}

=pod

---++ postRenderingHandler($text)

It does the same as  commonTagsHandler,  except for the results 
replaced here isn't rendered by  TWiki. This means that no kind
of text format, autolink, variable expansion, etc is done. Some
components may need this.

=cut

sub postRenderingHandler {
	# my ($text) = @_;
	$_[0] =~ s/%MASON_PLUGIN_UNRENDERED{app="(.*?)" call_id="(.*?)"}%/&_result($1, $2)/ge;
}

=pod

---++ %<nop>MASON% handler

Collects information about each component call.

=cut

sub MASON_Handler
{
	my ($session, $params, $theTopic, $theWeb) = @_;
	_parseOptions();
	my $app = _getDefaultApp($theTopic, $theWeb);
	if ($params->{app} && ( (ref($app_list{$theWeb}) eq 'ARRAY' && 
		(scalar grep { $params->{app} eq $_ } @{ $app_list{$theWeb} }) == 0) || 
		! ref($app_list{$theWeb})) ) {
		return _addComp(undef);
	}
	my %args;
	if ($params->{args}) {
		my $safe = Safe->new();
		%args = $safe->reval($params->{args});
		return $@ if $@;
	}
	my $postRender = $params->{post_render} || TWiki::Func::expandCommonVariables('%MASON_POST_RENDERING%');
	if ($postRender eq '%MASON_POST_RENDERING%') {
		$postRender = TWiki::Func::getPluginPreferencesFlag('POST_RENDERING') || 0;
	}
	elsif ($postRender =~ /^yes$/i || $postRender =~ /^true$/i || $postRender == 1) {
		$postRender = 1;
	}
	else {
		$postRender = 0;
	}
	return _addComp($app, $params->{_DEFAULT}, $params->{args}, $postRender);
}

=pod

---++ %<nop>MASON_ENABLED_APPS% handler

Return the list of enabled applications for the current web.

=cut

sub MASON_ENABLED_APPS_Handler 
{
	# my (\%session, \%params, $theTopic, $theWeb) = @_;
	_parseOptions();
	return $app_list{$_[3]} ? join(", ", @{ $app_list{$_[3]} }) : '=none=';
}

=pod

---++ %<nop>MASON_APPS% handler

Return the list of all found applications, regardless permission.
To see a  list of  enabled  applications  in a  context  refer to
%MASON_ENABLED_APPS% handler.

=cut

sub MASON_APPS_Handler
{
	# my (\%session, \%params, $theTopic, $theWeb) = @_;
	_parseOptions();
	return join(", ", @found_apps) || '=none=';
}

=pod

---++ %<nop>MASON_APPS_TABLE% handler

Display a table with applications as columns headers and webs in rows
in such a way it's possible to view all applications enabled/disabled
for all webs.

=cut

sub MASON_APPS_TABLE_Handler
{
	# my (\%session, \%params, $theTopic, $theWeb) = @_;
	_parseOptions();
	my %table;
	foreach my $web (keys %app_list) {
		$table{$web}{$_} = 1 foreach @{$app_list{$web}};
	}
	my $str = "\n| |  " . join("  |  ", map { '  *'.$_.'*  ' } @found_apps) . "  |\n";
	foreach my $web (sort keys %app_list) {
		$str .= "|  *$web*  |  ".join("  |  ",map { $table{$web}{$_} ? '%ICON{choice-yes}%' : '%ICON{choice-no}%' } @found_apps)."  |\n";
	}
	return scalar @found_apps ? $str : '=none=';
}

=pod

---++ %<nop>MASON_FAILED_APPS% handler

Display a table with error messages for applications, if any.

=cut

sub MASON_FAILED_APPS_Handler
{
	# my (\%session, \%params, $theTopic, $theWeb) = @_;
	_parseOptions();
	my $str = "";
	foreach my $app (keys %app_err_msg) {
		$app_err_msg{$app} =~ s/(\r?\n|\r)/\\$1/g;
		$app_err_msg{$app} =~ s/\|/!/g;
		$str .= "|  *$app*  |  ".$app_err_msg{$app}."  |\n";
	}
	return $str ? "\n" . $str : '=none=';
}

#============ INTERNAL FUNCTIONS ===============

=pod

---++ _parseOptions() 

Read ENABLE_*_IN config  options at  TWiki.MasonPlugin and
constructs  a table, %app_list, containing the list of all
enabled applications for each web:
$app_list{$web} = [app1, app2, app3, etc... ]
If the key $web doesn't exist or the value is undefined or 
a empty list, then no application is enabled for that web.
It also fills the  @found_apps. Any error messages are put
in $app_err_msg{$app} and $app is skiped.

=cut

sub _parseOptions 
{
	# Gets the time of last write to MasonPlugin topic, via the $meta object
	# returned by readTopic.
	my $time = ((TWiki::Func::readTopic('TWiki', 'MasonPlugin'))[0]->getRevisionInfo())[0];
	
	# If $time is different from $parsed, than the topic was modified and
	# preferences must be re-parsed.
	return if $parsed == $time;
	my $apps_dir = $TWiki::cfg{MasonPlugin}{ApplicationsDir}.'/';
	my @apps =  split /[,\s]+/, $TWiki::cfg{MasonPlugin}{Applications} || '';
	@found_apps = ();
	%app_err_msg = ();
	%app_list = ();
	my @todas_webs = ();
	foreach my $app (@apps) {
		eval { require $apps_dir.$app.'.pm'; };
		if ($@) {
			$app_err_msg{$app} = $@;
			next;
		}
		push @found_apps, $app;
		my %webs = ();
		foreach my $keyWord (split(/[,\s]+/, TWiki::Func::getPluginPreferencesValue('ENABLE_'.(uc $app).'_IN') || '')) {
			if ($keyWord eq 'ALL') {
				@todas_webs = TWiki::Func::getListOfWebs('user') unless scalar @todas_webs;
				$webs{$_} = 1 foreach @todas_webs;
			}
			elsif ($keyWord =~ /^-\s*(\S*)/) {
				delete $webs{$1};
			}
			else {
				$webs{$keyWord} = 1 if TWiki::Func::webExists($keyWord);
			}
		}
		push @{$app_list{$_}}, $app foreach keys %webs;
	}
	$parsed = $time;
}

=pod

---++ _getDefaultApp($theTopic, $theWeb) -> $app

Return  the  default app for the  context of $theWeb.$theTopic, 
which is  determined  by  the  %MASON_DEFAULT_APP% (topic/web), 
MASON_DEFAULT_APP variable at plugin or the application enabled 
if there is only one. undef otherwise (or if the especified one 
isn't enabled for $theWeb.$theTopic

=cut

sub _getDefaultApp 
{
	# my ($theTopic, $theWeb) = @_;
	_parseOptions();
	my $app = TWiki::Func::expandCommonVariables('%MASON_DEFAULT_APP%', $_[0], $_[1]);
	$app =	TWiki::Func::getPluginPreferencesValue('MASON_DEFAULT_APP') if $app eq '%MASON_DEFAULT_APP%';
	$app = $app_list{$_[1]}->[0] if !$app && @{$app_list{$_[1]}} == 1;
	return ($app ne '' && scalar grep { $_ eq $app } @{$app_list{$_[1]}}) ? $app : undef;
}

=pod

---++ _addComp($app, $comp, $args, $postRender) -> $text_
   * =$app= - Application
   * =$comp= - Component
   * =$args= - Arguments to be passed to =$comp=
   * =$postRender= - true if this call should be solved after rendering
   * =$text= - id text used later to replace the component call with the result

=cut

sub _addComp
{
	return "TWiki.MasonPlugin: couldn't figure out the application ". 
	       "or it is not enabled. Verify configuration at TWiki.MasonPlugin" unless $_[0];
	my $id = ref($comps{$_[0]}) eq 'ARRAY' ? scalar @{$comps{$_[0]}} : 0;
	push @{$comps{$_[0]}}, "<& $_[1], ".'%ARGS, '.$_[2]." &>";
	return '%MASON_PLUGIN_'.($_[3] ? 'UN' : '').'RENDERED{app="'.$_[0].'" call_id="'.$id.'"}%';
}

=pod

---++ _execComps()

For each found application within a topic create a macro-component 
based on all componente calls and execute it.

=cut

sub _execComps
{
	return if $executed;
	my $cgi = TWiki::Func::getCgiQuery();
	my %cgi_params = $cgi ? $cgi->Vars() : ();
	foreach (keys %cgi_params) {
		$cgi_params{$_} = [split "\0", $cgi_params{$_}] if $cgi_params{$_} =~ /\0/;
	}
	my %results;

APP: 
	foreach my $app (keys %comps) {
		my @interp_args;
		my ($interp, $outBuffer);
		eval {
			@interp_args = $app->interpArgs();
		};
		if ($@) {
			$results{$app}->[$_] = (split / at /, $@)[0] foreach 1 .. $#{$comps{$app}};
			next APP;
		}
		$interp = HTML::Mason::Interp->new(@interp_args, out_method => \$outBuffer);
		my $src = q|<%flags>
inherit => '/autohandler'
</%flags>
|;
		my $random = int(rand(2**30));
		$src .= join("\n\n===MasonPluginComponentDelimiter$random===\n\n", @{$comps{$app}});
		eval {
			my $comp = $interp->make_component(comp_source => $src);
			$interp->exec($comp, %cgi_params, masonShared => {});
		};
		@{$results{$app}} = split /\n*===MasonPluginComponentDelimiter$random===\n*/, $outBuffer || $@;
	}
	%comps = %results;
	$executed = 1;
}

=pod

---++ _result($app, $call_id) -> $text

Return the result os the call to component of application 
=$app= and id =$call_id=

=cut

sub _result
{
	_execComps() unless $executed;
	return $comps{$_[0]}->[$_[1]];
}

1;
